# semibake.py: Automatically assemble a half-baked knightship
# of any sufficiently large period (> 2400000-ish)
#
# Adam P. Goucher
# (c) 2014 Metaswitch Networks*
#
# * (It's in my contract of employment that any software that
#      I develop over this summer includes this message. :)

from time import sleep
import os
import golly as g
from glife import *
from string import *
import sys

# Class representing a trail of half-bakeries:
class Breadline:
    halfbakery = pattern("4b2o$3bo2bo$3bobo$b2obo$o2bo$obo$bo!")
    def __init__(self, ix, iy):
        self.p = pattern("")
        self.x = ix
        self.y = iy
    def vacuum(self, r):
        self.x -= r
        self.y += r
    def normal_HB(self):
        self.p += self.halfbakery(self.x,self.y)
        self.x += 1
        self.y += 1
    def perturbed_HB(self):
        self.p += self.halfbakery(self.x-1,self.y+1)
        self.x += 1
        self.y += 1

# Class representing the half-baked knightship itself:
class HBK:
    spacing = 24
    def __init__(self, w, k):
        self.width = w
        self.debt = -8*(k/47)
        self.lane = k
        self.retardation = 0
        self.rightdelay = k*2

        self.periphery = pattern("")
        
        self.AR = Breadline(0,0)
        self.AL = Breadline(37,13)
        self.AU = Breadline(21,21)

        self.BL = Breadline(110+w,83+w)
        self.BU = Breadline(94+w,91+w)

        self.CL = Breadline(137+w,113+w)
        self.CU = Breadline(121+w,121+w)
        self.CG = Breadline(104+w,127+w)

    def uniform_vacuum(self, r):
        self.AR.vacuum(r)
        self.AL.vacuum(r)
        self.AU.vacuum(r)

        self.BL.vacuum(r)
        self.BU.vacuum(r)

        self.CL.vacuum(r)
        self.CU.vacuum(r)
        self.CG.vacuum(r)

        self.retardation += 8*r

    def reset(self):
        self.debt += 8
        for i in xrange(8):
            self.AR.normal_HB()
            self.AL.normal_HB()
            self.AU.perturbed_HB()
            self.uniform_vacuum(self.spacing)
        self.retardation += 156
        self.lane -= 39

    def cancel_debt(self):
        for i in xrange(self.debt):
            self.BL.normal_HB()
            self.BU.perturbed_HB()

            self.CL.normal_HB()
            self.CU.perturbed_HB()
            self.CG.normal_HB()

            self.uniform_vacuum(self.spacing)
        self.debt = 0

    def shoot(self):
        self.AR.normal_HB()
        self.AL.normal_HB()
        self.AU.normal_HB()

        self.BL.normal_HB()
        self.BU.normal_HB()

        self.CL.normal_HB()
        self.CU.normal_HB()
        self.CG.normal_HB()

        self.uniform_vacuum(self.spacing)
        self.retardation += 43
        self.lane += 1

    def nonshoot(self):
        self.AR.normal_HB()
        self.AL.normal_HB()
        self.AU.perturbed_HB()

        self.BL.normal_HB()
        self.BU.perturbed_HB()

        self.CL.normal_HB()
        self.CU.perturbed_HB()
        self.CG.normal_HB()

        self.uniform_vacuum(self.spacing)
        self.retardation += 43
        self.lane += 1        

    def import_lanes(self,lanes):
        # -1000 corresponds to a forced reset; -10000-n gives a retardation of 8*n
        p = 0
        for l in lanes:
            if (l <= -10000):
                self.uniform_vacuum(-10000-l)
            else:
                if (l == -1000):
                    p -= 39
                    self.reset()
                else:
                    while (l < p):
                        p -= 39
                        self.reset()
                    while (p < l):
                        p += 1
                        self.nonshoot()
                    self.shoot()
                    p += 1

    def shoot_mwss(self, n, m):
        
        d = 2

        receptacle = pattern("")

        receptacle += pattern("2o$2o3$20b2o$20b2o15$2b2o$2b2o3$22b2o$22b2o!")(59, -23)
        receptacle += pattern("2o$2o53$66b2o$66b2o!")(3182, -3186)(n-180, 180-n)

        self.periphery += receptacle(171 + self.lane + m - n, -19 + self.lane - m + n)

        self.import_lanes([0, 1, 18+d, 19+d, 30+d, 31+d, 41+d, 45+d, 46+d, -2, 2, 8, 41+d, 42+d, 44+d, \
            50+d, 2, 4, 39+d, -1000, -1000, -1000, 3, 10, 29+d, 30+d, 31+d, 32+d, 33+d, \
            -1000, -1000, -1000, -3, 7, -10000-(n/2), 25+d, 26+d, 27+d, 30+d, 37+d, 45+d, \
            -1000, -1000, -1000, 16+d, 20+d, 26+d, 32+d, 38+d, \
            60, 61, -1000, 58, 62, 68, -1000, 62, 64, -1000, 63, 70, -1000, 57, 67])


    def shoot_mwss2(self, path, delay):

        if ((self.lane - path - 2389) % 2 == 1):
            self.reset()

        while ((delay - self.retardation - 4*(self.lane - path - 2389)) % 8 != 0):
            self.reset()
            for i in xrange(39):
                self.nonshoot()

        self.shoot_mwss(self.lane - path - 2839, (delay - self.retardation - 4*(self.lane - path - 2389))/8)

    def shoot_pi_even(self, n, m):

        receptacle = pattern("2o$2o19$2b2o$2b2o!")
        halfbakery = pattern("4b2o$3bo2bo$3bobo$b2obo$o2bo$obo$bo!")
        
        for i in xrange(n):
            receptacle += halfbakery(-15-10*i, 62+12*i)

        self.periphery += receptacle(142 + self.lane + m, 6 + self.lane - m)

        self.import_lanes([0, 1, 11, -2, 2, 6, 8, 13, 14, 8, 9, 11, 15, 8, 19, 12])

    def shoot_pi_even2(self, path, delay):

        while ((delay - self.retardation - 43*(path + (self.width - self.lane) + 23)) % 8 != 0):
            self.reset()
            for i in xrange(39):
                self.nonshoot()

        self.shoot_pi_even(path + (self.width - self.lane) + 23, (delay - self.retardation - 43*(path + (self.width - self.lane) + 23))/8)

    def shoot_pi_odd(self, n, m):

        receptacle = pattern("2o$2o29$3b2o$3b2o$34b2o$34bobo$35b2o4$16b2o$16b2o3$24b2o$24b2o!")
        halfbakery = pattern("4b2o$3bo2bo$3bobo$b2obo$o2bo$obo$bo!")

        for i in xrange(n):
            receptacle += halfbakery(3-10*i, 61+12*i)

        self.periphery += receptacle(131 + self.lane + m, 17 + self.lane - m)

        self.import_lanes([0, 1, 15, 27, 28, -2, 2, 8, 22, 25, 28, 2, 4, 8, -1, 3, 9])

    def shoot_pi_odd2(self, path, delay):

        while ((delay - self.retardation - 43*(path + (self.width - self.lane) + 5)) % 8 != 0):
            self.reset()
            for i in xrange(39):
                self.nonshoot()

        self.shoot_pi_odd(path + (self.width - self.lane) + 5, (delay - self.retardation - 43*(path + (self.width - self.lane) + 5))/8)
        
    def assemble(self):

        # Ensure everything is correctly timed for the cleanup mechanism:
        while (self.debt < 0):
            self.reset()
        self.uniform_vacuum(80)
        if (self.debt == 0):
            self.uniform_vacuum(160)
        self.cancel_debt()

        # Prepare an empty universe:
        g.new("Half-baked knightship")
        g.setrule("B3/S23")
        g.setalgo("QuickLife")
        
        # Draw the bread lines:
        self.AR.p.put(0,0)
        self.AL.p.put(0,0)
        self.AU.p.put(0,0)
        self.BL.p.put(0,0)
        self.BU.p.put(0,0)
        self.CL.p.put(0,0)
        self.CU.p.put(0,0)
        self.CG.p.put(0,0)
        self.periphery.put(0,0)

        # Draw the gliders:
        w = self.width
        h = self.rightdelay
        pattern("o$obo$2o!").put(55+w,-52-w)
        pattern("bo$o$3o!").put(73+w,-21-w)
        pattern("2bo$2o$b2o!").put(51+w,-7-w)
        pattern("bo$o$3o!").put(219+w+h,-24+w-h)
        pattern("2bo$2o$b2o!").put(197+w+h,-10+w-h)
        pattern("bo$o$3o!").put(224+w+h,28+w-h)
        pattern("2bo$2o$b2o!").put(202+w+h,42+w-h)
        pattern("obo$2o$bo!").put(177+w+h,57+w-h)

        # Add self-destructing machinery:
        pattern("""
184b2o$183bo2bo$183bobo$181b2obo$180bo2bo$180bobo$181bo6$174b2o$173bo
2bo$173bobo$171b2obo$170bo2bo$170bobo$171bo16$200b2o$199bo2bo$199bobo$
197b2obo$196bo2bo$196bobo$197bo2$184b2o$183bo2bo$183bobo$181b2obo$180b
o2bo$180bobo$181bo9$179b2o$178bo2bo$178bobo$176b2obo$175bo2bo$175bobo$
176bo6$169b2o$168bo2bo$168bobo$166b2obo$165bo2bo$165bobo$166bo6$159b2o
$158bo2bo$158bobo$156b2obo$155bo2bo$155bobo$156bo6$149b2o$148bo2bo$
148bobo$146b2obo$145bo2bo$145bobo$146bo5$124b2o$123bo2bo12b2o$123bobo
12bo2bo$121b2obo13bobo$120bo2bo12b2obo$120bobo12bo2bo$121bo13bobo$136b
o5$114b2o$113bo2bo12b2o$113bobo12bo2bo$111b2obo13bobo$110bo2bo12b2obo$
110bobo12bo2bo$111bo13bobo$126bo5$104b2o$103bo2bo12b2o$103bobo12bo2bo$
101b2obo13bobo$100bo2bo12b2obo$100bobo12bo2bo$101bo13bobo$116bo5$94b2o
$93bo2bo12b2o$93bobo12bo2bo$91b2obo13bobo$90bo2bo12b2obo$90bobo12bo2bo
$91bo13bobo$106bo5$84b2o$83bo2bo12b2o$83bobo12bo2bo$81b2obo13bobo$80bo
2bo12b2obo$80bobo12bo2bo$81bo13bobo$96bo5$74b2o$73bo2bo12b2o$73bobo12b
o2bo$71b2obo13bobo$70bo2bo12b2obo$70bobo12bo2bo$71bo13bobo$86bo5$64b2o
$63bo2bo12b2o$63bobo12bo2bo$61b2obo13bobo$60bo2bo12b2obo$60bobo12bo2bo
$61bo13bobo$76bo5$54b2o$53bo2bo12b2o$53bobo12bo2bo$51b2obo13bobo$50bo
2bo12b2obo$50bobo12bo2bo$51bo13bobo$66bo5$44b2o$43bo2bo$43bobo$41b2obo
$40bo2bo$40bobo$41bo6$34b2o$33bo2bo$33bobo$31b2obo$30bo2bo$30bobo$31bo
6$24b2o$23bo2bo$23bobo$21b2obo$20bo2bo$20bobo$21bo6$14b2o$13bo2bo$13bo
bo$11b2obo$10bo2bo$10bobo$11bo6$4b2o$3bo2bo$3bobo$b2obo$o2bo$obo$bo!
""").put(26+self.AR.x,-206+self.AR.y)

        pattern("""
63b2o$62bo2bo$62bobo$60b2obo$59bo2bo$59bobo$60bo2$47b2o$46bo2bo$46bobo
$44b2obo$43bo2bo$43bobo$44bo6$37b2o$36bo2bo$36bobo$34b2obo$33bo2bo$33b
obo$34bo19$75b2o$74bo2bo$74bobo$72b2obo$71bo2bo$71bobo$72bo2$59b2o$58b
o2bo$58bobo$56b2obo$55bo2bo$55bobo$56bo6$49b2o$48bo2bo$34b2o12bobo$33b
o2bo9b2obo$33bobo9bo2bo$31b2obo10bobo$30bo2bo12bo$30bobo$31bo6$24b2o$
23bo2bo$23bobo$21b2obo$20bo2bo$20bobo$21bo6$14b2o$13bo2bo$13bobo$11b2o
bo$10bo2bo$10bobo$11bo6$4b2o$3bo2bo$3bobo$b2obo$o2bo$obo$bo!
""").put(-101+self.CL.x,-15+self.CL.y)

        # Reposition the viewport to contain the pattern:
        g.fit()
        g.setalgo("HashLife")


def semibake(period):

    hbk = HBK(15040,7520)
    hbk.shoot_mwss2(0, period - 65526 - 107467)
    for i in xrange(80):
        hbk.nonshoot()
    hbk.shoot_mwss2(21, period - 65526 - 97 - 107467)
    for i in xrange(80):
        hbk.nonshoot()
    hbk.shoot_mwss2(25, period - 65526 - 26 - 107467)
    for i in xrange(80):
        hbk.nonshoot()
    hbk.shoot_pi_even2(0, period - 65526 - 2194)
    for i in xrange(80):
        hbk.nonshoot()
    hbk.shoot_pi_even2(-4, period - 65526 - 71 - 2194)
    for i in xrange(80):
        hbk.nonshoot()
    hbk.shoot_pi_odd2(0, period - 65526)
    for i in xrange(80):
        hbk.nonshoot()
    hbk.shoot_pi_odd2(-19, period - 65526 + 245)
    for i in xrange(80):
        hbk.nonshoot()
    hbk.shoot_pi_odd2(-23, period - 65526 + 174)
    hbk.assemble()


semibake(int(g.getstring("Enter period:", "2621440")))
